#' @rdname CST_RainFARM
#' @title RainFARM Stochastic Precipitation Downscaling of a CSTools object
#'
#' @author Jost von Hardenberg, ISAC-CNR, \email{j.vonhardenberg@isac.cnr.it}
#'
#' @description This function implements the RainFARM stochastic precipitation downscaling method and accepts a CSTools object in input.
#' Adapted for climate downscaling and including orographic correction.
#' References: Terzago, S. et al. (2018). NHESS 18(11), 2825–2840. http://doi.org/10.5194/nhess-18-2825-2018 ,
#' D'Onofrio et al. 2014, J of Hydrometeorology 15, 830-843; Rebora et. al 2006, JHM 7, 724.
#' Please notice: The function uses julia_setup from the JuliaCall library, 
#' which takes a long time only the first time it is called on a machine.
#'
#' @param s2d_data CSTools object containing the precipitation array to downscale
#' @param weights matrix (can be obtained using the \code{RFWeights} function (default 1. == none)
#' @param nf refinement factor for downscaling (the output resolution is increased by this factor)
#' @param slope prescribed or automatic spectral slope (default 0. == automatic)
#' @param kmin first wavenumber for spectral slope
#' @param nens number of ensemble members to produce
#' @param fglob logical to conseve global precipitation over the domain (default FALSE)
#' @param fsmooth logical to conserve precipitation with a smnoothing kernel (default TRUE)
#' @param time_dim string or character array with name(s) of dimension(s) (e.g. "ftime", "sdate", "member" ...) 
#' over which to compute spectral slopes. If a character array of dimension names is provided, the spectral slopes
#' will be computed as an average over all elements belonging to those dimensions.
#' If omitted one of c("ftime", "sdate", "time")  is searched and the first one with more than one element is chosen.
#' @param verbose logical for verbose output of the Julia package
#' @param drop_realization_dim logical to remove the realization dimension if nens==1 (default TRUE)
#'  
#' @return CST_RainFARM() returns a downscaled CSTools object.
#' @import multiApply
#' @export
#' @examples
#' #Example using CST_RainFARM for a CSTools object
#' nf <- 8   # Choose a downscaling by factor 8
#' exp <- 1 : (2 * 3 * 4 * 8 * 8)
#' dim(exp) <- c(dataset = 1, member = 2, sdate = 3, ftime = 4, lat = 8, lon = 8)
#' lon <- seq(10, 13.5, 0.5)
#' dim(lon) <- c(lon = length(lon))
#' lat <- seq(40, 43.5, 0.5)
#' dim(lat) <- c(lat = length(lat))
#' s2dv_data <- list(exp = exp, lon = lon, lat = lat)
#' # Create a test array of weights
#' ww <- array(1., dim = c(8 * nf, 8 * nf))
#' res_s2dv <- CST_RainFARM(s2dv_data, nf, ww, nens=3)
#' str(res_s2dv)
#' #List of 3
#' # $ exp: num [1:4, 1:64, 1:64, 1, 1:2, 1:3] 201 115 119 307 146 ...
#' # $ lon : num [1:64] 9.78 9.84 9.91 9.97 10.03 ...
#' # $ lat : num [1:64] 39.8 39.8 39.9 40 40 ...
#' dim(res_s2dv$exp)
#' dataset  member   sdate   ftime     lat     lon    realization
#'      1       2       3       4      64      64     3
#'

CST_RainFARM <- function(s2d_data, nf, weights = 1., slope = 0, kmin = 1,
                         nens = 1, fglob = FALSE, fsmooth = FALSE,
                         time_dim = NULL, verbose = FALSE,
                         drop_realization_dim = TRUE) {

  res <- RainFARM(s2d_data$exp, s2d_data$lon, s2d_data$lat,
                  nf, weights, nens, slope, kmin, fglob, fsmooth,
                  time_dim, lon_dim = "lon", lat_dim = "lat",
                  drop_realization_dim, verbose)

  s2d_data$exp <- res$data
  s2d_data$lon <- res$lon
  s2d_data$lat <- res$lat

  return(s2d_data)
}

#' @rdname RainFARM
#' @title RainFARM Stochastic Precipitation Downscaling (reduced version)
#' @author Jost von Hardenberg, ISAC-CNR, \email{j.vonhardenberg@isac.cnr.it}
#' @description This function implements the RainFARM stochastic precipitation downscaling method
#' and accepts an array with named dims ("lon", "lat") and a time dimension ("ftime", "sdate" or "time") in input.
#' Adapted for climate downscaling and including orographic correction.
#' References: Terzago, S. et al. (2018). NHESS 18(11), 2825–2840. http://doi.org/10.5194/nhess-18-2825-2018 ,
#' D'Onofrio et al. 2014, J of Hydrometeorology 15, 830-843; Rebora et. al 2006, JHM 7, 724.
#' Please notice: The function uses julia_setup from the JuliaCall library, 
#' which takes a long time only the first time it is called on a machine.
#' @param data precipitation array to downscale
#' @param lon vector or array of longitudes
#' @param lat vector or array of latitudes
#' @param weights matrix (can be obtained using the \code{RFWeights} function (default 1. == none)
#' @param nf refinement factor for downscaling (the output resolution is increased by this factor)
#' @param slope prescribed or automatic spectral slope (default 0. == automatic)
#' @param kmin first wavenumber for spectral slope
#' @param nens number of ensemble members to produce
#' @param fglob logical to conseve global precipitation over the domain (default FALSE)
#' @param fsmooth logical to conserve precipitation with a smnoothing kernel (default TRUE)
#' @param time_dim string or character array with name(s) of time dimension(s) (e.g. "ftime", "sdate", "time" ...) 
#' over which to compute spectral slopes. If a character array of dimension names is provided, the spectral slopes
#' will be computed over all elements belonging to those dimensions.
#' If omitted one of c("ftime", "sdate", "time")  is searched and the first one with more than one element is chosen.
#' @param lon_dim name of lon dimension ("lon" by default).
#' @param lat_dim name of lat dimension ("lat" by default).
#' @param verbose logical for verbose output of the Julia package
#' @param drop_realization_dim logical to remove the ensemble dimension if nens==1
#' @return RainFARM() returns a list containing the fine-scale longitudes, latitudes and the sequence of \code{nens} downscaled fields.
#' @import JuliaCall
#' @import multiApply
#' @export
#' @examples
#' # Example for the 'reduced' RainFARM function 
#' nf <- 8   # Choose a downscaling by factor 8
#' nens <- 3 # Number of ensemble members
#' # create a test array with dimension 8x8 and 20 timesteps
#' # or provide your own read from a netcdf file
#' pr <- rnorm(8*8*20)
#' dim(pr) <- c(lon = 8, lat = 8, ftime = 20)
#' lon_mat <- seq(10,13.5,0.5) # could also be a 2d matrix
#' lat_mat <- seq(40,43.5,0.5)
#' # Create a test array of weights
#' ww <- array(1., dim = c(8 * nf, 8 * nf))
#' # ... or create proper weights using an external fine-scale climatology file
#' #     Specify a weightsfn filename if you wish to save the weights
#' ww <- RFWeights("./worldclim.nc", nf, lon=lon_mat, lat=lat_mat, weightsfn="", fsmooth = TRUE)
#' # downscale using weights (ww=1. means do not use weights)
#' res <- RainFARM(pr, lon_mat, lat_mat, nf, fsmooth = TRUE, fglob = FALSE, weights = ww, nens = 2, verbose = TRUE)
#' str(res)
#' #List of 3
#' # $ data: num [1:3, 1:20, 1:64, 1:64] 0.186 0.212 0.138 3.748 0.679 ...
#' # $ lon : num [1:64] 9.78 9.84 9.91 9.97 10.03 ...
#' # $ lat : num [1:64] 39.8 39.8 39.9 40 40 …
#' dim(res$data)
#' #  lon         lat       ftime realization 
#' #   64          64          20           2 

RainFARM <- function(data, lon, lat, nf, weights = 1., nens = 1,
                     slope = 0, kmin = 1, fglob = FALSE, fsmooth = TRUE,
                     time_dim = NULL, lon_dim = "lon", lat_dim = "lat",
                     drop_realization_dim = TRUE, verbose = FALSE) {

  # Check/detect time_dim
  if (is.null(time_dim)) {
    time_dim_names <- c("ftime", "sdate", "time")
    time_dim_num <- which(time_dim_names %in% names(dim(data)))
    if (length(time_dim_num) > 0) {
    # Find time dimension with length > 1
      ilong <- which(dim(data)[time_dim_names[time_dim_num]] > 1)
      if (length(ilong) > 0) {
        time_dim <- time_dim_names[time_dim_num[ilong[1]]]
      } else {
        stop("No time dimension longer than one found.")
      }
    } else {
      stop("Could not automatically detect a target time dimension ",
           "in the provided data in 'data'.")
    }
    print(paste("Selected time dim: ", time_dim))
  }

  # Check availability of JuliaCall
  if (!("JuliaCall" %in% rownames(installed.packages()))) {
    stop("Package 'JuliaCall' needs to be installed in order ",
         "to run RainFARM.")
  }

  # Perform common calls
  JuliaCall::julia_setup(useRCall = FALSE)
  julia_library("RainFARM")

  r <- julia_call("lon_lat_fine", lon, lat, as.integer(nf),
                  need_return = "R")
  lon_f <- r[[1]]
  lat_f <- r[[2]]

  # reorder and group time_dim together at the end
  cdim0 <- dim(data)
  imask <- names(cdim0) %in% time_dim
  data <- aperm(data,c(which(!imask), which(imask)))
  cdim <- dim(data)
  ind <- 1:length(which(!imask))
  # compact (multiply) time_dim dimensions
  dim(data) <- c(cdim[ind], rainfarm_samples = prod(cdim[-ind]))  
 
  # Repeatedly apply .RainFARM
  result <- Apply(data, c(lon_dim, lat_dim, "rainfarm_samples"), .RainFARM,
                  weights, nf, nens, slope, kmin,
                  fglob, fsmooth, verbose)$output1

  # Reorder as it was in data before apply + realization dim at end
  iorder=sapply(c(names(dim(data)), "realization"), grep, names(dim(result)))
  result=aperm(result, iorder)  
  # Expand back to original dims + realization dim at end
  dim(result) <- c(dim(result)[ind], cdim[-ind],
                   dim(result)[length(dim(result))])
  # Reorder as it was in data + realization dim at end
  iorder=sapply(c(names(cdim0), "realization"), grep, names(dim(result)))
  result=aperm(result, iorder)  
  if (drop_realization_dim && nens == 1) {
    dim(result) <- dim(result)[-which(names(dim(result)) == "realization")[1]]
  }

  return(list(data = result, lon = lon_f, lat = lat_f))
}

#' Atomic RainFARM
#' @param pr precipitation array to downscale with dims (lon, lat, time)
#' @param weights matrix (can be obtained using the \code{RFWeights} function (default 1. == none)
#' @param nf refinement factor for downscaling (the output resolution is increased by this factor)
#' @param slope prescribed or automatic spectral slope (default 0. == automatic)
#' @param kmin first wavenumber for automatic spectral slope
#' @param nens number of ensemble members to produce
#' @param fglob logical to conseve global precipitation over the domain (default FALSE)
#' @param fsmooth logical to conserve precipitation with a smnoothing kernel (default TRUE)
#' @param verbose logical for verbose output of the Julia package
#' @return .RainFARM returns a downscaled array with dims (lon, lat, time, realization)
#' @noRd

.RainFARM <- function(pr, weights, nf, nens, slope, kmin,
                      fglob, fsmooth, verbose) {

  if (slope == 0) {
    fxp <- julia_call("fft3d", pr, need_return = "R")[[1]]
    sx <- julia_call("fitslopex", fxp, kmin = as.integer(kmin),
                     need_return = "R")
  } else {
    sx <- slope
  }

  result_dims <- c(dim(pr)[1] * nf, dim(pr)[2] * nf, dim(pr)[3], realization = nens)
  r <- array(dim = result_dims)

  for (i in 1:nens) {
    r[, , , i] <- julia_call("rainfarm", pr, sx, as.integer(nf), weights,
                             fglob = fglob, fsmooth = fsmooth,
                             verbose = verbose, need_return = "R")
  }
  return(r)
}
